// ContinuousTime.cpp: implementation of the ContinuousTime class.
//
//////////////////////////////////////////////////////////////////////

#include "ContinuousTime.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
// #define new DEBUG_NEW
#endif

#include <assert.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

ContinuousTime::ContinuousTime(SpinLattice *S) : SpinDynamics(S)
{
	SetProbFlip();

	SetupLOOKLOC();
}


ContinuousTime::~ContinuousTime()
{
	delete [] LOOK[0];
	delete [] LOOK;
}



void ContinuousTime::SetProbFlip()
{
	double T = temperature;
	if (temperature == 0) temperature = 1e-99;	// Avoid divide by zero

	for (int s=0; s<=1; s++)	// Loop over spin +1, -1
	{
		for (int nUp=0; nUp<=4; nUp++)	// Loop over neighbors up
		{
			double dE = -2 * (2*(nUp-2) + magneticField);
			double exponent = -(2*s-1)*dE/temperature;
			if (exponent < -353.5) probFlip[s][nUp]=1.0;	// Avoid overflow
			else if (exponent > 353.5) probFlip[s][nUp]=0.0;
			else probFlip[s][nUp] = 1./(1.+exp(exponent));
		}
	}

}

void ContinuousTime::SetTemperature(double temperature)
{
	SpinDynamics::SetTemperature(temperature);
	SetProbFlip();
}

void ContinuousTime::SetMagneticField(double magneticField)
{
	SpinDynamics::SetMagneticField(magneticField);
	SetProbFlip();
}

void ContinuousTime::FixLOOKLOC(Point &changed, int dNUp, int ds)	// ds = dNUp = +-1
{
	// Given a neighbor of i, j has flipped so nUpNew-nUpOld = dNUp
	// or i,j has flipped so ds = sNew - sOld
	// move ij to correct LOC list
	// update LOOK list
		
		// move neighbors from list with old nUp to list with new nUp

		int sOld = (S->Get(changed.i,changed.j)+1)/2;
		int sNew = sOld + ds;
		int nUpOld = S->GetNeighborsUp(changed.i,changed.j);
		int nUpNew = nUpOld + dNUp;

		// Append onto new list
		LOC[sNew][nUpNew].push_back(changed);

		// Move end of old list onto old value: pop end of old list off

		Point lastInList = LOC[sOld][nUpOld].back();
		LOC[sOld][nUpOld][LOOK[changed.i][changed.j]] = lastInList;	// Third index is vector index!
		LOOK[lastInList.i][lastInList.j] = LOOK[changed.i][changed.j];	// lastInList now stored where changed used to be
		LOC[sOld][nUpOld].pop_back();	// Shorten old list by one
		LOOK[changed.i][changed.j]=LOC[sNew][nUpNew].size()-1;	// LOOK contains the location in LOC where ij is stored

}

void ContinuousTime::ResetLattice(SpinLattice * S)
{
	SpinDynamics::ResetLattice(S);
	delete [] LOOK[0];
	delete [] LOOK;
	for (int s = 0; s<2; s++)
	{
		for (int nUp=0; nUp<=4; nUp++)	LOC[s][nUp].erase(LOC[s][nUp].begin(),LOC[s][nUp].end());
	}

	SetProbFlip();
	SetupLOOKLOC();
}	

void ContinuousTime::SetupLOOKLOC()
{
	LOOK = new int *[S->GetHeight()];
	LOOK[0] = new int[S->GetHeight()*S->GetWidth()];

	for (int i=0; i<S->GetHeight(); i++)
	{
		LOOK[i] = &(LOOK[0][i*S->GetWidth()]);	// Pointer to first element in row
	}

	for (i=0; i<S->GetHeight(); i++)
	{
		for (int j=0; j<S->GetWidth(); j++)
		{
			int s = (S->Get(i,j)+1)/2;
			int nUp = S->GetNeighborsUp(i,j);
			Point ij;
			ij.i = i;
			ij.j = j;
			LOC[s][nUp].push_back(ij);
			LOOK[i][j] = LOC[s][nUp].size()-1;
		}
	}

}

double ContinuousTime::Sweep()
{
	double timeSweep = 0;
	int s, nUp;		// Define outside loops so value accessible when loop breaks

	while (timeSweep <= 1)
	{
		double totalRate = 0;
		for (s=0; s<=1; s++)	// Loop over spin +1, -1
		{
			for (nUp=0; nUp<=4; nUp++)	//  Loop over neighbors up
			{
				totalRate += LOC[s][nUp].size() * probFlip[s][nUp];
			}
		}

		double R = rand.uniform()*totalRate;

		// Find type of next spin to flip

		double Qsum = 0;
		s = -1;
		
		// As far as I can see, this is the most elegant way
		// I'll see in class if anyone has a better method!
		// Basically, by incrementing s and nUp before Qsum, and 
		// checking before incrementing again, loop ends with 
		// correct s and nUp

		while (s<1 && R>=Qsum)	// Loop over spin +1, -1
		{
			s++;
			nUp=-1;
			while (nUp < 4 && R >= Qsum)	// Loop over neighbors up
			{
				nUp++;
				Qsum += LOC[s][nUp].size()*probFlip[s][nUp];
			}
		}

		// s and nUp should point to the right environment class for the spin to flip now

		// pick random spin with that environment

		int toFlipIndex = rand.discrete(0,LOC[s][nUp].size()-1); // the index in the LOC array for the spin
		Point toFlip = LOC[s][nUp][toFlipIndex];	// the i,j position of the spin

		// Change in number of spins up

		int dNUp;
		if (s==0) dNUp = 1;
		else dNUp = -1;	

		// move neighbors from list with old nUp to list with new nUp

		Point next;
		next.i = toFlip.i+1;
		next.j = toFlip.j;
		if (next.i ==S->GetHeight()) next.i=0;
		FixLOOKLOC(next,dNUp,0);
		
		next.i = toFlip.i-1;
		if (next.i==-1) next.i = S->GetHeight()-1;
		FixLOOKLOC(next,dNUp,0);

		next.i=toFlip.i;
		next.j=toFlip.j+1;
		if (next.j==S->GetWidth()) next.j=0;
		FixLOOKLOC(next,dNUp,0);

		next.j=toFlip.j-1;
		if (next.j==-1) next.j = S->GetWidth()-1;
		FixLOOKLOC(next,dNUp,0);

		// Flip spin, shift lists, and change LOOK

		FixLOOKLOC(toFlip,0,dNUp);
		S->Set(toFlip.i, toFlip.j, dNUp);

		// update timeSweep by correct amount

		double store = rand.uniform();	// Avoid log(0)!
		if (store==0.0) store = 1.0;

		timeSweep += -log(store)/totalRate;

	}

	return timeSweep;
}
